/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

#define FE_IMO_DEBUG		FALSE

using namespace fe;

void ImportOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("format")="auto";
	catalog<String>("format","label")="Format";
	catalog<String>("format","choice:0")="auto";
	catalog<String>("format","label:0")="Auto";
	catalog<String>("format","choice:1")="xml";
	catalog<String>("format","label:1")="XML";
	catalog<String>("format","choice:2")="abc";
	catalog<String>("format","label:2")="Alembic";

	catalog<String>("loadfile")="";
	catalog<String>("loadfile","label")="Load File";
	catalog<String>("loadfile","suggest")="filename";

	catalog<bool>("filterJoints")="";
	catalog<String>("filterJoints","label")="Filter Joints";
	catalog<bool>("filterJoints","joined")=true;

	catalog<String>("filterFile")="";
	catalog<String>("filterFile","label")="Filter File";
	catalog<String>("filterFile","suggest")="filename";
	catalog<String>("filterFile","enabler")="filterJoints";
	catalog<bool>("filterFile","joined")=true;

	catalog<String>("filterSpec")="";
	catalog<String>("filterSpec","label")="Spec";
	catalog<String>("filterSpec","enabler")="filterJoints";

	catalog<Real>("frame")=0.0;
	catalog<Real>("frame","high")=1e3;
	catalog<Real>("frame","max")=1e6;
	catalog<String>("frame","label")="Frame";

	catalog<String>("fallbackFrame")="100";
	catalog<String>("fallbackFrame","label")="Fallback Frame";

	catalog<bool>("cacheFrames")=true;
	catalog<String>("cacheFrames","label")="Cache Frames";

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spScope=registry()->create("Scope");
}

void ImportOp::handle(Record& a_rSignal)
{
#if FE_IMO_DEBUG
	feLog("ImportOp::handle node \"%s\"\n",name().c_str());
#endif

	//* TODO need "eval" for Houdini
	const String loadfile=catalogOrDefault<String>("loadfile","eval",
			catalog<String>("loadfile"));
	const BWORD filterJoints=catalog<bool>("filterJoints");
	const String filterFile=filterJoints?
			catalogOrDefault<String>("filterFile","eval",
			catalog<String>("filterFile")): "";
	const String filterSpec=catalog<String>("filterSpec");
	const BWORD cacheFrames=catalog<bool>("cacheFrames");

	if(loadfile.empty())
	{
		return;
	}

	String format=catalog<String>("format");
	if(format=="auto")
	{
		format=String(strrchr(loadfile.c_str(),'.')).prechop(".");
	}

	if(m_loadedFormat!=format)
	{
		m_loadedFormat=format;

		m_spLoadAccessible=NULL;
		m_frameCache.clear();
	}
	if(m_loadedFile!=loadfile)
	{
		m_loadedFile=loadfile;

		m_spLoadAccessible=NULL;
		m_frameCache.clear();
	}
	if(m_loadedFilter!=filterFile)
	{
		m_loadedFilter=filterFile;

		m_spLoadAccessible=NULL;
		m_frameCache.clear();
	}
	if(m_loadedSpec!=filterSpec)
	{
		m_loadedSpec=filterSpec;

		m_spLoadAccessible=NULL;
		m_frameCache.clear();
	}

	const BWORD quiet=TRUE;

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	if(m_spLoadAccessible.isNull())
	{
		m_spLoadAccessible=registry()->create(
				"SurfaceAccessibleI.*.*."+format,quiet);
	}
	if(m_spLoadAccessible.isNull())
	{
		m_spLoadAccessible=registry()->create("SurfaceAccessibleI.*.*.xml");
	}
	if(m_spLoadAccessible.isNull())
	{
		m_spLoadAccessible=registry()->create("SurfaceAccessibleI");
	}
	if(m_spLoadAccessible.isNull())
	{
		catalog<String>("error")+="no format drivers found";
		return;
	}

#if FE_IMO_DEBUG
	feLog("ImportOp::handle using \"%s\"\n",
			m_spLoadAccessible->name().c_str());
#endif

	sp<StringFilterI> spStringFilter;
	if(!filterFile.empty())
	{
		const String filterFormat=
			String(strrchr(filterFile.c_str(),'.')).prechop(".");

		spStringFilter=registry()->create("StringFilterI.*.*."+filterFormat);
		if(spStringFilter.isValid())
		{
			spStringFilter->configure(filterSpec);
			if(!spStringFilter->load(filterFile))
			{
				spStringFilter=NULL;
			}
		}

		if(!spStringFilter.isValid())
		{
			catalog<String>("warning")+=
					"failed to access filter file \""+filterFile+"\"";
		}
	}

	Real frame=catalog<Real>("frame");
	if(frame<startFrame(a_rSignal) || frame>endFrame(a_rSignal))
	{
		frame=currentFrame(a_rSignal);
	}

	const I32 cacheIndex=frame*1000;	//* TODO precision param
	if(cacheFrames)
	{
		sp<SurfaceAccessibleI> spCacheFrame=m_frameCache[cacheIndex];
		if(spCacheFrame.isValid())
		{
#if FE_IMO_DEBUG
			feLog("ImportOp::handle use cache %d\n",cacheIndex);
#endif

			spOutputAccessible->copy(spCacheFrame);
			return;
		}
	}
	else
	{
		m_frameCache.clear();
	}

	const BWORD substitute=loadfile.match(".*[$]F.*");

	const I32 intFrame=frame;
	const String intText=print(intFrame);
	const String realText=print(frame);
	String modfile=loadfile.substitute("$F",print(intFrame))
			.substitute("$FF",print(frame));

#if FE_IMO_DEBUG
	feLog("ImportOp::handle start %.6G end %.6G frame %.6G -> %.6G\n",
			startFrame(a_rSignal),endFrame(a_rSignal),
			catalog<Real>("frame"),frame);
	feLog("ImportOp::handle sub %d load \"%s\"\n",substitute,modfile.c_str());
#endif

	m_spLoadAccessible->bind(m_spScope);
	U32 jointCount=0;
	if(!m_spLoadAccessible->load(modfile,frame,spStringFilter))
	{
		feLog("ImportOp::handle"
				" node \"%s\" failed to load:\n  \"%s\"\n  filter \"%s\"\n",
				name().c_str(),modfile.c_str(),filterFile.c_str());
	}
	else
	{
		jointCount=m_spLoadAccessible->count(SurfaceAccessibleI::e_primitive);
		if(!jointCount)
		{
			feLog("ImportOp::handle"
					" node \"%s\" parsed zero joints:\n"
					"  \"%s\"\n  filter \"%s\"\n",
					name().c_str(),modfile.c_str(),filterFile.c_str());
		}
	}

	if(!jointCount)
	{
		if(!substitute)
		{
			catalog<String>("error")+="load failed (no substitution);";
		}
		else
		{
			const String fallbackFrame=
					catalog<String>("fallbackFrame","eval");

			modfile=loadfile.substitute("$F",fallbackFrame)
					.substitute("$FF",fallbackFrame);

			if(m_spLoadAccessible->load(modfile,frame,spStringFilter) &&
					m_spLoadAccessible->count(SurfaceAccessibleI::e_primitive))
			{
				feLog("  -> using fallback frame \"%s\"\n",
						fallbackFrame.c_str());

				catalog<String>("warning")+=
						"substituting fallback frame \""+fallbackFrame+"\"";
			}
			else
			{
				feLog("  -> also failed using fallback frame \"%s\"\n",
						fallbackFrame.c_str());

				const Real current=currentFrame(a_rSignal);
				const String currentText=print(current);

				modfile=loadfile.substitute("$F",print(I32(current)))
						.substitute("$FF",currentText);

				if(m_spLoadAccessible->load(modfile,frame,spStringFilter) &&
						m_spLoadAccessible->count(
						SurfaceAccessibleI::e_primitive))
				{
					feLog("  -> using current frame \"%s\"\n",
							currentText.c_str());

					catalog<String>("warning")+=
							"substituting current frame \""+currentText+"\"";
				}
				else
				{
					feLog("  -> also failed using current frame \"%s\"\n",
							currentText.c_str());

					catalog<String>("error")+="load failed;";
					m_spLoadAccessible=NULL;
				}
			}
		}
	}

	if(m_spLoadAccessible.isValid())
	{
//		sp<SurfaceAccessibleI> spOutputAccessible;
//		if(!accessOutput(spOutputAccessible,a_rSignal)) return;

		spOutputAccessible->copy(m_spLoadAccessible);
		if(cacheFrames)
		{
#if FE_IMO_DEBUG
			feLog("ImportOp::handle store cache %d\n",cacheIndex);
#endif
			m_frameCache[cacheIndex]=m_spLoadAccessible;
			m_spLoadAccessible=NULL;
		}
	}
}
