/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __platform_h__
#define __platform_h__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <typeinfo>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <set>
#include <stack>
#include <queue>

#include "fe/config.h"

#include "platform/define.h"

// hash_map is not 'standard', so we have some hoops
#if defined(GNU_UNORDERED_MAP)
#include <unordered_map>
#include <unordered_set>
#elif defined(HASH_MAP_IS_MAP)
namespace ext
{
	template<class k, class v, class h, class e>
	class hash_map : public std::map< k,v > { };
	template<class k, class h, class e>
	class hash_set : public std::set< k > { };
}
#elif defined(GNU_STL_HASH_MAP)
#include <ext/hash_map>
#include <ext/hash_set>
namespace ext = __gnu_cxx;
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
#include <hash_map>
#include <hash_set>
#else
#include <ext/hash_map>
#include <ext/hash_set>
#endif

#if defined(GNU_UNORDERED_MAP)
namespace fe
{
	template<class k, class v, class h, class e>
	class HashMap : public std::unordered_map< k,v,h,e > { };
	template<class k, class h, class e>
	class HashSet : public std::unordered_set< k,h,e > { };
}
#else
namespace fe
{
	template<class k, class v, class h, class e>
	class HashMap : public ext::hash_map< k,v,h,e > { };
	template<class k, class h, class e>
	class HashSet : public ext::hash_set< k,h,e > { };
}
#endif

#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64

	#include <CRTDbg.h>
	#define NOMINMAX
	#if FE_COMPILER!=FE_DMC
	#include <winsock2.h>
	#endif
	#include <windows.h>	//* causes wacky struct errors (???)
	#include <mmsystem.h>
	#include <direct.h>

	#define mkdir _mkdir

	#ifndef WIN32_LEAN_AND_MEAN
		#define WIN32_LEAN_AND_MEAN
	#endif

	#define FE_MESSAGE_LENGTH_MAX					512

#elif FE_OS==FE_LINUX
	#include <sys/time.h>
	#include <sys/stat.h>
	#include <sys/types.h>
	#include <limits.h>
	#include <unistd.h>
	#include <errno.h>
	#include <netinet/in.h>
	#include <dlfcn.h>
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <netdb.h>

	#define FE_MESSAGE_LENGTH_MAX					512

#elif FE_OS==FE_OSX
	#include <sys/time.h>
	#include <sys/stat.h>
	#include <sys/types.h>
	#include <unistd.h>
	#include <errno.h>
	#include <netinet/in.h>
	#include <dlfcn.h>
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <netdb.h>
	// and I thought MS was presumptuous...check() is a macro for OSX
	#include <AssertMacros.h>
	#ifdef check
	#undef check
	#endif

	#define FE_MESSAGE_LENGTH_MAX					512

#elif FE_OS==FE_CYGWIN
	#include <sys/time.h>
	#include <sys/types.h>
	#include <unistd.h>
	#include <errno.h>
	#include <netinet/in.h>
	#include <dlfcn.h>
	#include <sys/socket.h>
	#include <arpa/inet.h>
	#include <netdb.h>

	#define FE_MESSAGE_LENGTH_MAX					512
#endif

#if FE_COMPILER==FE_GNU && FE_OS==FE_OSX
#ifndef GCC_HASCLASSVISIBILITY
#define GCC_HASCLASSVISIBILITY
#endif
#endif

#if FE_COMPILER==FE_MICROSOFT
	#define FE_DL_EXPORT	__declspec(dllexport)
	#define FE_DL_IMPORT	__declspec(dllimport)
	#define FE_DL_PUBLIC
	#define FE_DL_LOCAL
#elif FE_COMPILER==FE_GNU && defined(GCC_HASCLASSVISIBILITY)
	//* NOTE GCC manual specifies empty FE_DL_IMPORT
	#define FE_DL_EXPORT	__attribute__ ((visibility("default")))
	#define FE_DL_IMPORT
	#define FE_DL_PUBLIC	FE_DL_EXPORT
	#define FE_DL_LOCAL		__attribute__ ((visibility("hidden")))
#else
	#define FE_DL_EXPORT
	#define FE_DL_IMPORT
	#define FE_DL_PUBLIC
	#define FE_DL_LOCAL
#endif

#if FE_COMPILER==FE_GNU
	#include <new>
#elif FE_COMPILER==FE_MICROSOFT
	//* DISABLE:  'this' : used in base member initializer list
	#pragma warning( disable : 4355 )
	//* DISABLE:  dll-interface complaints
	#pragma warning( disable : 4275 )
	#pragma warning( disable : 4251 )
	#include <new.h>
#else
	#include <new.h>
#endif


#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	#include <mbstring.h>
#else
	#include <ctype.h>
#endif

#ifdef FE_BOOST_EXCEPTIONS
#include "boost/exception/diagnostic_information.hpp"
#endif

#include "boost/dynamic_bitset.hpp"
typedef boost::dynamic_bitset<>	t_bitset;

//* fePlatform is a static lib; FE_PLATFORM_PORT should not be defined.
//* You should not be importing or exporting anything from fePlatform.
//* In Win32, you can get multiple symbols and MSVCRT lib problems.
//#ifdef MODULE_platform
//#define FE_PLATFORM_PORT FE_DL_EXPORT
//#else
//#define FE_PLATFORM_PORT FE_DL_IMPORT
//#endif

#if FE_CODEGEN>FE_DEBUG
#define FEASSERT(condition)
#elif FE_OS==FE_WIN32 || FE_OS==FE_WIN64
#define FEASSERT(condition) _ASSERT(condition)
#else
#define FEASSERT(condition) assert(condition)
#endif

#include "platform/datatypes.h"
#include "platform/error.h"
#include "platform/Mutex.h"
#include "platform/memory.h"
#include "platform/Regex.h"
#include "platform/count.h"
#include "platform/String.h"
#include "platform/Thread.h"
#include "platform/backtrace.h"
#include "platform/debug.h"
#include "platform/System.h"
#include "platform/Result.h"
#include "platform/Exception.h"
#include "platform/UnitTest.h"
#include "platform/DL_Loader.h"
//#include "platform/SystemTicker.h"

#endif /* __platform_h__ */

